/**
 ** Streams PPP module
 **/

#include "f_module.h"
#include "primitives.h"
#include <sys/types.h>
#include <sys/time.h>
#include "f_signal.h"
#include "f_malloc.h"
#include <sys/param.h>
#include <sys/sysmacros.h>
#include "streams.h"
#include <sys/stropts.h>
#ifdef DONT_ADDERROR
#include "f_user.h"
#endif
#include <sys/errno.h>
#include "streamlib.h"
#include "strppp.h"
#include "isdn_proto.h"


/*
 * Standard Streams stuff.
 */
static struct module_info ppp_minfo =
{
		0, "ppp", 0, INFPSZ, 4096, 2048
};

static qf_open ppp_open;
static qf_close ppp_close;
static qf_put ppp_rput,ppp_wput;
static qf_srv ppp_rsrv,ppp_wsrv;

static struct qinit ppp_rinit =
{
		ppp_rput, ppp_rsrv, ppp_open, ppp_close, NULL, &ppp_minfo, NULL
};

static struct qinit ppp_winit =
{
		ppp_wput, ppp_wsrv, NULL, NULL, NULL, &ppp_minfo, NULL
};

struct streamtab pppinfo =
{&ppp_rinit, &ppp_winit, NULL, NULL};

struct ppp_ {
	char flags;
#define PPP_COMPRESS 02
	ushort_t offset;
};


/* Streams code to open the driver. */
static int
ppp_open (queue_t * q, dev_t dev, int flag, int sflag
#ifdef DO_ADDERROR
		,int *err
#endif
)
{
	register struct ppp_ *ppp;

	if (q->q_ptr) {
		return 0;
	}
	ppp = malloc(sizeof(*ppp));
	if(ppp == NULL)
		return OPENFAIL;
	memset(ppp,0,sizeof(*ppp));

	WR (q)->q_ptr = (char *) ppp;
	q->q_ptr = (char *) ppp;

	MORE_USE;
	return 0;
}

/* Streams code to close the driver. */
static void
ppp_close (queue_t * q, int dummy)
{
	register struct ppp_ *ppp;

	ppp = (struct ppp_ *) q->q_ptr;

	flushq (q, FLUSHALL);
	flushq (WR (q), FLUSHALL);
	if (0)
		printf ("PPP driver closed.\n");
	free(ppp);
	LESS_USE;
	return;
}


static void
ppp_proto (queue_t * q, mblk_t * mp, char isdown)
{
	register struct ppp_ *ppp = (struct ppp_ *) q->q_ptr;
	streamchar *origmp = mp->b_rptr;
	ushort_t id;

    if (m_getid (mp, &id) != 0) {
        mp->b_rptr = origmp;
        putnext (q, mp);
        return;
    }
	switch (id) {
	default:
		break;
	case PROTO_OFFSET: {
			long z;
			if (m_geti (mp, &z) != 0)
				goto err;
			if (z < 0 || z >= 1024)
				goto err;
			if(isdown) {
				ppp->offset = z;
				z += 1;
				freemsg(mp);
				if((mp = allocb(10,BPRI_MED)) != NULL) {
					m_putid(mp,PROTO_OFFSET);
					m_puti(mp,z);
					mp->b_datap->db_type = MSG_PROTO;
					origmp = NULL;
				}
			}
		} break;
	case PROTO_MODULE:
		if (strnamecmp (q, mp)) {	/* Config information for me. */

			while (mp != NULL && m_getsx (mp, &id) == 0) {
				switch (id) {
				default:
					break;
				case PPP_COMPR:
					ppp->flags |= PPP_COMPRESS;
					break;
				case PPP_NOCOMPR:
					ppp->flags &= ~PPP_COMPRESS;
					break;
				}
			}
			if (mp != NULL) {
				mp->b_rptr = origmp;
				m_reply(q,mp,0);
				mp = NULL;
			}
		}
	}
	if(mp != NULL)  {
		if (origmp != NULL)
			mp->b_rptr = origmp;
		putnext(q,mp);
	}
	return;
  err:
	mp->b_rptr = origmp;
	m_reply (q, mp, EINVAL);
}


/* Streams code to write data. */
static void
ppp_wput (queue_t * q, mblk_t * mp)
{
	switch (mp->b_datap->db_type) {
	case MSG_PROTO:
	CASE_DATA
		putq (q, mp);
		break;
	case M_FLUSH:
		if (*mp->b_rptr & FLUSHW)
			flushq (q, FLUSHDATA);
		putnext (q, mp);
		break;
	default:
		putnext (q, mp);
		break;
	}
	return;
}

/* Streams code to scan the write queue. */
static void
ppp_wsrv (queue_t * q)
{
	register struct ppp_ *ppp = (struct ppp_ *) q->q_ptr;
	mblk_t *mp;

	while ((mp = getq (q)) != NULL) {
		switch (mp->b_datap->db_type) {
		case MSG_PROTO:
			ppp_proto (q, mp, 1);
			break;
		CASE_DATA {
				if (!canput (q->q_next)) {
					putbq (q, mp);
					return;
				}
				if((ppp->flags & PPP_COMPRESS) && *mp->b_rptr == 0)
					mp->b_rptr++;

				putnext (q, mp);
			} break;
		case M_FLUSH:
			if (*mp->b_rptr & FLUSHW)
				flushq (q, FLUSHDATA);
			/* FALL THRU */
		default:
			if (mp->b_datap->db_type > QPCTL || canput (q->q_next)) {
				putnext (q, mp);
				continue;
			} else {
				putbq (q, mp);
				return;
			}
		}
	}
	return;
}

/* Streams code to read data. */
static void
ppp_rput (queue_t * q, mblk_t * mp)
{
	switch (mp->b_datap->db_type) {

	case M_FLUSH:
		if (*mp->b_rptr & FLUSHR) {
			flushq (q, FLUSHDATA);
		}
		putnext (q, mp);		  /* send it along too */
		break;
	case MSG_PROTO:
	CASE_DATA
		putq (q, mp);			  /* queue it for my service routine */
		break;

	default:
		putq (q, mp);
	}
	return;
}

/* Streams code to scan the read queue. */
static void
ppp_rsrv (queue_t * q)
{
	mblk_t *mp;
	register struct ppp_ *ppp = (struct ppp_ *) q->q_ptr;

	while ((mp = getq (q)) != NULL) {
		switch (mp->b_datap->db_type) {
		case MSG_PROTO:
			ppp_proto (q, mp, 0);
			break;
		CASE_DATA {
				mblk_t *mh;

				if (!canput (q->q_next)) {
					putbq (q, mp);
					return;
				}
				if(*mp->b_rptr & 1) {
					if (/* XXX */ 0 || mp->b_datap->db_ref > 1 ||
							mp->b_datap->db_base > mp->b_rptr - 2) {
						mh = allocb(1+ppp->offset,BPRI_LO);
						if(mh == NULL) {
							putbqf(q,mp);
							return;
						}
						mh->b_datap->db_type = mp->b_datap->db_type;
						mh->b_rptr += ppp->offset+1;
						mh->b_wptr += ppp->offset+1;
						linkb (mh, mp);
						mp = mh;
					}
					*--mp->b_rptr = 0x00;
				}

				putnext (q, mp);
			} break;
		default:
			if (mp->b_datap->db_type > QPCTL || canput (q->q_next)) {
				putnext (q, mp);
				continue;
			} else {
				putbq (q, mp);
				return;
			}
		}
	}
	return;
}


#ifdef MODULE
static int do_init_module(void)
{
	return register_strmod(&pppinfo);
}

static int do_exit_module(void)
{
	return unregister_strmod(&pppinfo);
}
#endif
